import logging
from contextlib import contextmanager
from typing import List, Tuple

from _pytest.logging import LogCaptureHandler


@contextmanager
def catch_logs(level: int, logger: logging.Logger) -> LogCaptureHandler:
    """Context manager that sets the level for capturing of logs.

    After the end of the 'with' statement the level is restored to its original value.

    :param level: The level.
    :param logger: The logger to update.
    """
    handler = LogCaptureHandler()
    orig_level = logger.level
    logger.setLevel(level)
    logger.addHandler(handler)
    try:
        yield handler
    finally:
        logger.setLevel(orig_level)
        logger.removeHandler(handler)


def records_to_tuples(records: List[logging.LogRecord]) -> List[Tuple[str, int, str]]:
    """A list of a stripped down log records intended for use in assertion comparison.

    :param records: A list of LogRecord objects.
    :returns: A list of tuples, where each tuple has the format (logger_name, log_level, message)
    """
    return [(r.name, r.levelno, r.getMessage()) for r in records]



import pytest
import requests
import requests_mock

from nldcsc.http_apis.base_class.api_base_class import ApiBaseClass


class HttpApi(ApiBaseClass):
    def __init__(self, baseurl: str, **kwargs):
        super().__init__(baseurl, **kwargs)

    def get_dummy_endpoint(self):
        resource = "dummy"

        return self.call(self.methods.GET, resource=resource)

    def get_dummy_endpoint_url_param(self, data=None):
        resource = "dummy"

        return self.call(self.methods.GET, resource=resource, data=data)

    def get_dummy_response_endpoint(self):
        resource = "dummy"
        data = {"data": "data"}

        return self.call(
            self.methods.GET, resource=resource, return_response_object=True, data=data
        )

    def post_str_dummy(self):
        resource = "dummy"
        data = "data"

        return self.call(self.methods.POST, resource=resource, data=data)

    def put_str_dummy(self):
        resource = "dummy"
        data = "data"

        return self.call(self.methods.PUT, resource=resource, data=data)

    def patch_dict_dummy(self):
        resource = "dummy"
        data = {"data": "data"}

        return self.call(self.methods.PATCH, resource=resource, data=data)

    def delete_dict_dummy(self):
        resource = "dummy"

        return self.call(self.methods.DELETE, resource=resource)

    def post_invalid_data_dummy(self):
        resource = "dummy"
        data = 1

        return self.call(self.methods.POST, resource=resource, data=data)

    def post_unserializable_data_dummy(self):
        class DummyClass:
            pass

        resource = "dummy"
        data = {"data": DummyClass()}

        return self.call(self.methods.POST, resource=resource, data=data)


@pytest.fixture
def http_api():
    ha = HttpApi(baseurl="http://localhost:8000", user_agent="HTTP_API")

    yield ha


@pytest.fixture
def http_path_api():
    ha = HttpApi(
        baseurl="http://localhost:8000",
        user_agent="HTTP_API",
        api_path="api",
        verify=True,
    )

    yield ha


class TestHttpApis:
    def test_headers(self, http_api):
        assert http_api.headers == {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "HTTP_API",
        }

        http_api.set_header_field("Testing", "Custom header")

        assert http_api.headers["Testing"] == "Custom header"

        http_api.del_header_field("Testing")

        assert http_api.headers == {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "HTTP_API",
        }

        http_api.clear_headers()

        assert http_api.headers is None

        http_api.reset_headers()

        assert http_api.headers == {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "HTTP_API",
        }

    def test_paths(self, http_path_api, http_api):
        assert http_path_api.baseurl == "http://localhost:8000"

        assert http_path_api.api_path == "api"

        assert (
            http_path_api._build_url("resource") == "http://localhost:8000/api/resource"
        )

        assert http_api._build_url("resource") == "http://localhost:8000/resource"

    def test_dummy_calls(self, http_api):
        with pytest.raises(requests.ConnectionError):
            http_api.get_dummy_endpoint()

        with pytest.raises(requests.ConnectionError):
            http_api.get_dummy_endpoint_url_param()

        with pytest.raises(requests.ConnectionError):
            http_api.post_str_dummy()

        with pytest.raises(requests.ConnectionError):
            http_api.put_str_dummy()

        with pytest.raises(requests.ConnectionError):
            http_api.patch_dict_dummy()

        with pytest.raises(requests.ConnectionError):
            http_api.delete_dict_dummy()

    def test_session_calls(self, http_api):
        with requests_mock.Mocker() as m:
            m.get(
                "http://localhost:8000/dummy", text='{"data": "data"}', status_code=404
            )

            with pytest.raises(requests.ConnectionError):
                d = http_api.get_dummy_endpoint()

                assert d == {"data": "data"}

        with requests_mock.Mocker() as m:
            m.get(
                "http://localhost:8000/dummy?id=1&id2=10",
                json={"id": "1", "id2": 10, "result": "success"},
                status_code=200,
            )

            d = http_api.get_dummy_endpoint_url_param(data={"id": 1, "id2": 10})

            assert d == {"id": "1", "id2": 10, "result": "success"}

        with requests_mock.Mocker() as m:
            m.get(
                "http://localhost:8000/dummy", text='{"data": "data"}', status_code=200
            )

            d = http_api.get_dummy_endpoint_url_param()

            assert d == {"data": "data"}

        with requests_mock.Mocker() as m:
            m.get(
                "http://localhost:8000/dummy", text='{"data": "data"}', status_code=200
            )

            d = http_api.get_dummy_response_endpoint()

            assert isinstance(d, requests.Response)

        with requests_mock.Mocker() as m:
            m.get(
                "http://localhost:8000/dummy",
                text="just string data",
                status_code=200,
                headers={"content-type": "text-plain"},
            )

            d = http_api.get_dummy_endpoint()

            assert isinstance(d, requests.Response)

        with requests_mock.Mocker() as m:
            m.get(
                "http://localhost:8000/dummy",
                text="just string data",
                status_code=200,
                headers={"content-type": "text/plain"},
            )

            d = http_api.get_dummy_endpoint()

            assert isinstance(d, str)
            assert d == "just string data"

        with requests_mock.Mocker() as m:
            m.get(
                "http://localhost:8000/dummy",
                text="just string data",
                status_code=200,
            )

            http_api.get_dummy_endpoint()
        with pytest.raises(requests.ConnectionError):
            http_api.delete_dict_dummy()

        with requests_mock.Mocker() as m:
            m.post(
                "http://localhost:8000/dummy",
                text="just string data",
                status_code=200,
            )

            with pytest.raises(TypeError):
                http_api.post_invalid_data_dummy()

        with requests_mock.Mocker() as m:
            m.post(
                "http://localhost:8000/dummy",
                text="just string data",
                status_code=200,
            )

            with pytest.raises(TypeError):
                http_api.post_unserializable_data_dummy()



from sqlalchemy import JSON, TEXT, BigInteger, String
from sqlalchemy.orm import DeclarativeBase, registry
from .custom_types import char_UUID
from .annotations import (
    str_16,
    str_30,
    str_50,
    str_64,
    str_100,
    str_128,
    str_256,
    str_512,
    str_text,
    list_json,
    dict_json,
    big_int_pk,
    uuid,
    uuid_pk,
)


class ModelBase(DeclarativeBase):
    registry = registry(
        type_annotation_map={
            str_16: String(16),
            str_30: String(30),
            str_50: String(50),
            str_64: String(64),
            str_100: String(100),
            str_128: String(128),
            str_256: String(256),
            str_512: String(512),
            str_text: TEXT,
            dict_json: JSON,
            list_json: JSON,
            big_int_pk: BigInteger,
            uuid: char_UUID,
            uuid_pk: char_UUID,
        },
    )



import contextlib
import logging
from collections import namedtuple
from typing import Any

import httpx
from httpx import (
    Client,
    AsyncClient,
    HTTPTransport,
    AsyncHTTPTransport,
    HTTPError,
    Response,
)
from nldcsc.loggers.app_logger import AppLogger

logging.setLoggerClass(AppLogger)


class HttpxBaseClass(object):
    def __init__(
        self,
        baseurl: str,
        api_path: str = None,
        proxies: dict = None,
        user_agent: str = "HttpxBaseClass",
        use_async_client: bool = True,
        **kwargs,
    ):
        """
        The Generic api caller handles all communication towards an api resource.
        """
        self.logger = logging.getLogger(self.__class__.__name__)

        if "verify" not in kwargs:
            self.verify = False
        else:
            self.verify = kwargs.pop("verify")

        if "follow_redirects" not in kwargs:
            self.follow_redirects = True
        else:
            self.follow_redirects = kwargs.pop("follow_redirects")

        if "http2" not in kwargs:
            self.http2 = True
        else:
            self.http2 = kwargs.pop("http2")

        self.baseurl = baseurl
        self.api_path = api_path
        self.proxies = proxies
        self.user_agent = user_agent
        self.use_async_client = use_async_client
        self.kwargs = kwargs

        self.methods = namedtuple("methods", "GET POST DELETE PUT PATCH")(
            "GET", "POST", "DELETE", "PUT", "PATCH"
        )

        self.myheaders = self.__default_headers

    def __repr__(self) -> str:
        """return a string representation of the obj GenericApi"""
        return f"<< {self.__class__.__name__}:{self.baseurl} >>"

    def _build_url(self, resource: str) -> str:
        """
        Internal method to build an url to use when executing commands
        """
        if self.api_path is None:
            return f"{self.baseurl}/{resource}"
        else:
            return f"{self.baseurl}/{self.api_path}/{resource}"

    @property
    def __default_headers(self) -> dict:
        """
        Property to return the default headers
        """
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": f"{self.user_agent}",
        }

    @property
    def headers(self) -> dict:
        """
        Property to return the current headers
        """

        return self.myheaders

    def set_header_field(self, field: str, value: str) -> dict:
        """
        Method to add a header and set its value
        """
        try:
            self.myheaders[field] = value
            return self.myheaders
        except TypeError:
            self.myheaders = {field: value}
            return self.myheaders

    def del_header_field(self, field: str) -> dict:
        """
        Method to delete a header field
        """

        self.myheaders.pop(field)

        return self.myheaders

    def reset_headers(self):
        """
        Method to reset the headers to the default values
        """

        self.myheaders = self.__default_headers

    def clear_headers(self):
        """
        Method to clear the headers and the myheaders to None
        """
        self.myheaders = None

    def create_sync_client(self, **kwargs) -> Client:
        transport = HTTPTransport(retries=3, http2=self.http2)
        client = httpx.Client(
            verify=self.verify,
            proxies=self.proxies,
            follow_redirects=self.follow_redirects,
            transport=transport,
            http2=self.http2,
            **kwargs,
        )

        return client

    def create_async_client(self, **kwargs) -> AsyncClient:
        transport = AsyncHTTPTransport(retries=3, http2=self.http2)
        client = httpx.AsyncClient(
            verify=self.verify,
            proxies=self.proxies,
            follow_redirects=self.follow_redirects,
            transport=transport,
            http2=self.http2,
            **kwargs,
        )

        return client

    def get_client(self) -> Client | AsyncClient:
        if self.use_async_client:
            return self.create_async_client(**self.kwargs)
        else:
            return self.create_sync_client(**self.kwargs)

    @contextlib.contextmanager
    def get_cm_client(self) -> Client | AsyncClient:
        client = self.get_client()

        try:
            yield client
        except Exception as err:
            self.logger.warning(f"Error requesting client context manager: {err}")
            self.logger.exception(err)
        finally:
            client.close()

    def call(
        self,
        method: str = None,
        resources: str | list = None,
        data: dict = None,
        timeout: int = 60,
        stream: bool = False,
        generic_request: bool = False,
    ) -> Response | str | Any:
        """
        Method for requesting free format api resources
        """
        try:
            if stream:
                request_api_resource = {
                    "headers": self.myheaders,
                    "timeout": timeout,
                }

                with self.get_client() as client:
                    if isinstance(resources, str):
                        with client.stream(
                            method, resources, **request_api_resource
                        ) as response_stream:
                            # returning complete body; but other iter_* methods might be better suited for
                            # specific cases
                            return response_stream.read()
                    else:
                        raise TypeError(
                            "For streaming responses; resources must be a string"
                        )
            else:
                with self.get_client() as client:
                    if isinstance(resources, list):
                        results = []
                        for resource in resources:
                            results.append(
                                self._s_connect(
                                    method=method,
                                    resource=resource,
                                    client=client,
                                    data=data,
                                    timeout=timeout,
                                    generic_request=generic_request,
                                )
                            )
                    else:
                        resource = resources
                        results = self._s_connect(
                            method=method,
                            resource=resource,
                            client=client,
                            data=data,
                            timeout=timeout,
                            generic_request=generic_request,
                        )
                    return results
        except HTTPError:
            raise
        except Exception:
            raise

    async def a_call(
        self,
        method: str = None,
        resources: str | list = None,
        data: dict = None,
        timeout: int = 60,
        stream: bool = False,
        generic_request: bool = False,
    ) -> Response | str | Any:
        """
        Method for requesting free format api resources
        """
        try:
            if stream:
                request_api_resource = {
                    "headers": self.myheaders,
                    "timeout": timeout,
                }

                async with self.get_client() as client:
                    if isinstance(resources, str):
                        with client.stream(
                            method, resources, **request_api_resource
                        ) as response_stream:
                            # returning complete body; but other iter_* methods might be better suited for
                            # specific cases
                            return await response_stream.read()
                    else:
                        raise TypeError(
                            "For streaming responses; resources must be a string"
                        )
            else:
                async with self.get_client() as client:
                    if isinstance(resources, list):
                        results = []
                        for resource in resources:
                            results.append(
                                await self._a_connect(
                                    method=method,
                                    resource=resource,
                                    client=client,
                                    data=data,
                                    timeout=timeout,
                                    generic_request=generic_request,
                                )
                            )
                    else:
                        resource = resources
                        results = await self._a_connect(
                            method=method,
                            resource=resource,
                            client=client,
                            data=data,
                            timeout=timeout,
                            generic_request=generic_request,
                        )
                    return results
        except HTTPError:
            raise
        except Exception:
            raise

    def _s_connect(
        self,
        method: str,
        resource: str,
        client: Client,
        data: dict | bytes = None,
        timeout: int = 60,
        generic_request: bool = False,
    ) -> Response | str | Any:
        """
        Send a request
        """

        request_api_resource = {
            "headers": self.myheaders,
            "timeout": timeout,
        }

        if data is not None:
            if method in [self.methods.DELETE, self.methods.GET]:
                if not isinstance(data, bytes):
                    raise TypeError(
                        f"'data' body (data == dict) type for {method} is not supported; use content (data == bytes) "
                        f"instead. Or set generic_request to 'True'"
                    )
                request_api_resource["content"] = data
            else:
                request_api_resource["data"] = data

        try:
            if generic_request:
                r = client.request(
                    method, self._build_url(resource), **request_api_resource
                )
            else:
                r = getattr(client, method.lower())(
                    self._build_url(resource), **request_api_resource
                )

            return r
        except HTTPError:
            raise
        except Exception as err:
            raise Exception(err)

    async def _a_connect(
        self,
        method: str,
        resource: str,
        client: AsyncClient,
        data: dict | bytes = None,
        timeout: int = 60,
        generic_request: bool = False,
    ) -> Response | str | Any:
        """
        Send a request
        """
        request_api_resource = {
            "headers": self.myheaders,
            "timeout": timeout,
        }

        if data is not None:
            if method in [self.methods.DELETE, self.methods.GET]:
                if not isinstance(data, bytes):
                    raise TypeError(
                        f"'data' body (data == dict) type for {method} is not supported; use content (data == bytes) "
                        f"instead. Or set generic_request to 'True'"
                    )
                request_api_resource["content"] = data
            else:
                request_api_resource["data"] = data

        try:
            if generic_request:
                r = await client.request(
                    method, self._build_url(resource), **request_api_resource
                )
            else:
                r = await getattr(client, method.lower())(
                    self._build_url(resource), **request_api_resource
                )

            return r
        except HTTPError:
            raise
        except Exception as err:
            raise Exception(err)



import ast
import collections
import json
import os
import secrets
import string
from json import JSONDecodeError


def getenv_bool(name: str, default: str = "False"):
    raw = os.getenv(name, default).title()
    try:
        the_bool = ast.literal_eval(raw)

        if not isinstance(the_bool, bool):
            raise ValueError
    except ValueError:
        raise

    return the_bool


def getenv_list(name: str, default: list = None):
    if default is None:
        default = []

    raw = os.getenv(name, default)

    if not isinstance(raw, list):
        try:
            the_list = json.loads(raw.replace("\n", ""))
            return the_list
        except JSONDecodeError:
            raise

    return default


def getenv_dict(name: str, default: dict = None):
    if default is None:
        default = {}

    raw = os.getenv(name, default)

    if not isinstance(raw, dict):
        try:
            the_dict = json.loads(raw)
            return the_dict
        except JSONDecodeError:
            raise

    return default


def str2bool(v):
    return v.lower() in ("yes", "true", "1")


def generate_random_password():
    # define the alphabet
    letters = string.ascii_letters
    digits = string.digits

    alphabet = letters + digits

    # fix password length
    pwd_length = 32

    # generate a password string
    pwd = ""
    for i in range(pwd_length):
        pwd += "".join(secrets.choice(alphabet))

    return pwd


def reverse_from_named_tuple(
    n_tuple: collections.namedtuple, index: int | str, lowercase_output: bool = False
) -> str:
    n_list = [
        x.lower() if lowercase_output else x
        for x in n_tuple.__dir__()
        if not x.startswith("_") and x not in ["index", "count"]
    ]

    n_rev_types = {getattr(n_tuple, x.upper()): x for x in n_list}

    try:
        return n_rev_types[index]
    except KeyError:
        raise KeyError(f"The requested index does not exist! Choices are {n_rev_types}")



import os
from urllib.parse import urlparse

import redis

from nldcsc.generic.utils import getenv_dict


class FlaskRedis(object):
    def __init__(
        self, app=None, init_standalone: bool = False, redis_url: str = None, **kwargs
    ):
        if app and init_standalone:
            raise Exception("App must be None when 'init_standalone' is set to True")

        self._redis_client = None
        self.kwargs = kwargs
        self.redis_url = (
            redis_url
            if redis_url is not None
            else os.getenv("REDIS_URL", "redis://redis:6379/")
        )

        self.redis_cache_db = int(os.getenv("REDIS_CACHE_DB", 0))
        self.redis_kwargs = getenv_dict("REDIS_KWARGS", None)

        if app is not None or init_standalone:
            self.init_app(app)

    def init_app(self, app, **kwargs):
        self.kwargs.update(kwargs)

        if self.redis_kwargs is not None:
            self.kwargs.update(self.redis_kwargs)

        parsed_url = urlparse(self.redis_url)

        self._redis_client = redis.Redis(
            host=str(parsed_url.hostname),
            port=int(parsed_url.port),
            db=int(self.redis_cache_db),
            **self.kwargs,
        )

        if app is not None:
            app.redis_client = self

    def __repr__(self):
        return "<< FlaskRedis >>"

    @property
    def redis_client(self) -> redis.Redis:
        return self._redis_client

    @property
    def set(self):
        return self._redis_client.set

    @property
    def delete(self):
        return self._redis_client.delete

    @property
    def get(self):
        return self._redis_client.get
